/*       +------------------------------------+
 *       | Inspire Internet Relay Chat Daemon |
 *       +------------------------------------+
 *
 *  InspIRCd: (C) 2002-2008 InspIRCd Development Team
 * See: http://www.inspircd.org/wiki/index.php/Credits
 *
 * This program is free but copyrighted software; see
 *            the file COPYING for details.
 *
 * ---------------------------------------------------
 */

#include "inspircd.h"

/* $ModDesc: Provides support for the /WATCH command */


/*
 * Okay, it's nice that this was documented and all, but I at least understood very little
 * of it, so I'm going to attempt to explain the data structures in here a bit more.
 *
 * For efficiency, many data structures are kept.
 *
 * The first is a global list `watchentries':
 *	hash_map<irc::string, std::deque<User*> >
 *
 * That is, if nick 'w00t' is being watched by user pointer 'Brain' and 'Om', <w00t, (Brain, Om)>
 * will be in the watchentries list.
 *
 * The second is that each user has a per-user data structure attached to their user record via Extensible:
 *	std::map<irc::string, std::string> watchlist;
 * So, in the above example with w00t watched by Brain and Om, we'd have:
 * 	Brain-
 * 	      `- w00t
 * 	Om-
 * 	   `- w00t
 *
 * Hopefully this helps any brave soul that ventures into this file other than me. :-)
 *		-- w00t (mar 30, 2008)
 */


/* This module has been refactored to provide a very efficient (in terms of cpu time)
 * implementation of /WATCH.
 *
 * To improve the efficiency of watch, many lists are kept. The first primary list is
 * a hash_map of who's being watched by who. For example:
 *
 * KEY: Brain   --->  Watched by:  Boo, w00t, Om
 * KEY: Boo     --->  Watched by:  Brain, w00t
 * 
 * This is used when we want to tell all the users that are watching someone that
 * they are now available or no longer available. For example, if the hash was
 * populated as shown above, then when Brain signs on, messages are sent to Boo, w00t
 * and Om by reading their 'watched by' list. When this occurs, their online status
 * in each of these users lists (see below) is also updated.
 *
 * Each user also has a seperate (smaller) map attached to their User whilst they
 * have any watch entries, which is managed by class Extensible. When they add or remove
 * a watch entry from their list, it is inserted here, as well as the main list being
 * maintained. This map also contains the user's online status. For users that are
 * offline, the key points at an empty string, and for users that are online, the key
 * points at a string containing "users-ident users-host users-signon-time". This is
 * stored in this manner so that we don't have to FindUser() to fetch this info, the
 * users signon can populate the field for us.
 *
 * For example, going again on the example above, this would be w00t's watchlist:
 *
 * KEY: Boo    --->  Status: "Boo brains.sexy.babe 535342348"
 * KEY: Brain  --->  Status: ""
 *
 * In this list we can see that Boo is online, and Brain is offline. We can then
 * use this list for 'WATCH L', and 'WATCH S' can be implemented as a combination
 * of the above two data structures, with minimum CPU penalty for doing so.
 *
 * In short, the least efficient this ever gets is O(n), and thats only because
 * there are parts that *must* loop (e.g. telling all users that are watching a
 * nick that the user online), however this is a *major* improvement over the
 * 1.0 implementation, which in places had O(n^n) and worse in it, because this
 * implementation scales based upon the sizes of the watch entries, whereas the
 * old system would scale (or not as the case may be) according to the total number
 * of users using WATCH.
 */

/*
 * Before you start screaming, this definition is only used here, so moving it to a header is pointless.
 * Yes, it's horrid. Blame cl for being different. -- w00t
 */
#ifdef WINDOWS
typedef nspace::hash_map<irc::string, std::deque<User*>, nspace::hash_compare<irc::string, std::less<irc::string> > > watchentries;
#else
typedef nspace::hash_map<irc::string, std::deque<User*>, nspace::hash<irc::string> > watchentries;
#endif
typedef std::map<irc::string, std::string> watchlist;

/* Who's watching each nickname.
 * NOTE: We do NOT iterate this to display a user's WATCH list!
 * See the comments above!
 */
watchentries* whos_watching_me;

class CommandSVSWatch : public Command
{
 public:
	CommandSVSWatch (InspIRCd* Instance) : Command(Instance,"SVSWATCH", 0, 2)
	{
		this->source = "m_watch.so";
		syntax = "<target> [C|L|S]|[+|-<nick>]";
		TRANSLATE3(TR_NICK, TR_TEXT, TR_END); /* we watch for a nick. not a UID. */
	}

	CmdResult Handle (const char* const* parameters, int pcnt, User *user)
	{
		if (!ServerInstance->ULine(user->server))
			return CMD_FAILURE;
			
		User *u = ServerInstance->FindNick(parameters[0]);
		if (!u)
			return CMD_FAILURE;
			
		if (IS_LOCAL(u))
		{
			ServerInstance->Parser->CallHandler("WATCH", &parameters[1], 1, u);
		}
		
		return CMD_SUCCESS;
	}
};

/** Handle /WATCH
 */
class CommandWatch : public Command
{
	unsigned int& MAX_WATCH;
 public:
	CmdResult remove_watch(User* user, const char* nick)
	{
		// removing an item from the list
		if (!ServerInstance->IsNick(nick))
		{
			user->WriteNumeric(942, "%s %s :Invalid nickname", user->nick, nick);
			return CMD_FAILURE;
		}

		watchlist* wl;
		if (user->GetExt("watchlist", wl))
		{
			/* Yup, is on my list */
			watchlist::iterator n = wl->find(nick);

			if (!wl)
				return CMD_FAILURE;

			if (n != wl->end())
			{
				if (!n->second.empty())
					user->WriteNumeric(602, "%s %s %s :stopped watching", user->nick, n->first.c_str(), n->second.c_str());
				else
					user->WriteNumeric(602, "%s %s * * 0 :stopped watching", user->nick, nick);

				wl->erase(n);
			}

			if (!wl->size())
			{
				user->Shrink("watchlist");
				delete wl;
			}

			watchentries::iterator x = whos_watching_me->find(nick);
			if (x != whos_watching_me->end())
			{
				/* People are watching this user, am i one of them? */
				std::deque<User*>::iterator n2 = std::find(x->second.begin(), x->second.end(), user);
				if (n2 != x->second.end())
					/* I'm no longer watching you... */
					x->second.erase(n2);

				if (!x->second.size())
					whos_watching_me->erase(nick);
			}
		}

		/* This might seem confusing, but we return CMD_FAILURE
		 * to indicate that this message shouldnt be routed across
		 * the network to other linked servers.
		 */
		return CMD_FAILURE;
	}

	CmdResult add_watch(User* user, const char* nick)
	{
		if (!ServerInstance->IsNick(nick))
		{
			user->WriteNumeric(942, "%s %s :Invalid nickname",user->nick,nick);
			return CMD_FAILURE;
		}

		watchlist* wl;
		if (!user->GetExt("watchlist", wl))
		{
			wl = new watchlist();
			user->Extend("watchlist", wl);
		}

		if (wl->size() == MAX_WATCH)
		{
			user->WriteNumeric(512, "%s %s :Too many WATCH entries", user->nick, nick);
			return CMD_FAILURE;
		}

		watchlist::iterator n = wl->find(nick);
		if (n == wl->end())
		{
			/* Don't already have the user on my watch list, proceed */
			watchentries::iterator x = whos_watching_me->find(nick);
			if (x != whos_watching_me->end())
			{
				/* People are watching this user, add myself */
				x->second.push_back(user);
			}
			else
			{
				std::deque<User*> newlist;
				newlist.push_back(user);
				(*(whos_watching_me))[nick] = newlist;
			}

			User* target = ServerInstance->FindNick(nick);
			if (target)
			{
				if (target->Visibility && !target->Visibility->VisibleTo(user))
				{
					(*wl)[nick] = "";
					user->WriteNumeric(605, "%s %s * * 0 :is offline",user->nick, nick);
					return CMD_FAILURE;
				}

				(*wl)[nick] = std::string(target->ident).append(" ").append(target->dhost).append(" ").append(ConvToStr(target->age));
				user->WriteNumeric(604, "%s %s %s :is online",user->nick, nick, (*wl)[nick].c_str());
				if (IS_AWAY(target))
				{
					user->WriteNumeric(609, "%s %s %s %s %ld :is away", user->nick, target->nick, target->ident, target->dhost, target->awaytime);
				}
			}
			else
			{
				(*wl)[nick] = "";
				user->WriteNumeric(605, "%s %s * * 0 :is offline",user->nick, nick);
			}
		}

		return CMD_FAILURE;
	}

	CommandWatch (InspIRCd* Instance, unsigned int &maxwatch) : Command(Instance,"WATCH",0,0), MAX_WATCH(maxwatch)
	{
		this->source = "m_watch.so";
		syntax = "[C|L|S]|[+|-<nick>]";
		TRANSLATE2(TR_TEXT, TR_END); /* we watch for a nick. not a UID. */
	}

	CmdResult Handle (const char* const* parameters, int pcnt, User *user)
	{
		if (!pcnt)
		{
			watchlist* wl;
			if (user->GetExt("watchlist", wl))
			{
				for (watchlist::iterator q = wl->begin(); q != wl->end(); q++)
				{
					if (!q->second.empty())
						user->WriteNumeric(604, "%s %s %s :is online", user->nick, q->first.c_str(), q->second.c_str());
				}
			}
			user->WriteNumeric(607, "%s :End of WATCH list",user->nick);
		}
		else if (pcnt > 0)
		{
			for (int x = 0; x < pcnt; x++)
			{
				const char *nick = parameters[x];
				if (!strcasecmp(nick,"C"))
				{
					// watch clear
					watchlist* wl;
					if (user->GetExt("watchlist", wl))
					{
						for (watchlist::iterator i = wl->begin(); i != wl->end(); i++)
						{
							watchentries::iterator i2 = whos_watching_me->find(i->first);
							if (i2 != whos_watching_me->end())
							{
								/* People are watching this user, am i one of them? */
								std::deque<User*>::iterator n = std::find(i2->second.begin(), i2->second.end(), user);
								if (n != i2->second.end())
									/* I'm no longer watching you... */
									i2->second.erase(n);

								if (!i2->second.size())
									whos_watching_me->erase(user->nick);
							}
						}

						delete wl;
						user->Shrink("watchlist");
					}
				}
				else if (!strcasecmp(nick,"L"))
				{
					watchlist* wl;
					if (user->GetExt("watchlist", wl))
					{
						for (watchlist::iterator q = wl->begin(); q != wl->end(); q++)
						{
							if (!q->second.empty())
							{
								user->WriteNumeric(604, "%s %s %s :is online", user->nick, q->first.c_str(), q->second.c_str());
								User *targ = ServerInstance->FindNick(q->first.c_str());
								if (IS_AWAY(targ))
								{
									user->WriteNumeric(609, "%s %s %s %s %ld :is away", user->nick, targ->nick, targ->ident, targ->dhost, targ->awaytime);
								}
							}
							else
								user->WriteNumeric(605, "%s %s * * 0 :is offline", user->nick, q->first.c_str());
						}
					}
					user->WriteNumeric(607, "%s :End of WATCH list",user->nick);
				}
				else if (!strcasecmp(nick,"S"))
				{
					watchlist* wl;
					int you_have = 0;
					int youre_on = 0;
					std::string list;

					if (user->GetExt("watchlist", wl))
					{
						for (watchlist::iterator q = wl->begin(); q != wl->end(); q++)
							list.append(q->first.c_str()).append(" ");
						you_have = wl->size();
					}

					watchentries::iterator i2 = whos_watching_me->find(user->nick);
					if (i2 != whos_watching_me->end())
						youre_on = i2->second.size();

					user->WriteNumeric(603, "%s :You have %d and are on %d WATCH entries", user->nick, you_have, youre_on);
					user->WriteNumeric(606, "%s :%s",user->nick, list.c_str());
					user->WriteNumeric(607, "%s :End of WATCH S",user->nick);
				}
				else if (nick[0] == '-')
				{
					nick++;
					remove_watch(user, nick);
				}
				else if (nick[0] == '+')
				{
					nick++;
					add_watch(user, nick);
				}
			}
		}
		/* So that spanningtree doesnt pass the WATCH commands to the network! */
		return CMD_FAILURE;
	}
};

class Modulewatch : public Module
{
	CommandWatch* mycommand;
	CommandSVSWatch *sw;
	unsigned int maxwatch;
	
 public:
	Modulewatch(InspIRCd* Me)
		: Module(Me), maxwatch(32)
	{
		OnRehash(NULL, "");
		whos_watching_me = new watchentries();
		mycommand = new CommandWatch(ServerInstance, maxwatch);
		ServerInstance->AddCommand(mycommand);
		sw = new CommandSVSWatch(ServerInstance);
		ServerInstance->AddCommand(sw);
		Implementation eventlist[] = { I_OnRehash, I_OnGarbageCollect, I_OnCleanup, I_OnUserQuit, I_OnPostConnect, I_OnUserPostNick, I_On005Numeric, I_OnSetAway };
		ServerInstance->Modules->Attach(eventlist, this, 8);
	}

	virtual void OnRehash(User* user, const std::string &parameter)
	{
		ConfigReader Conf(ServerInstance);
		maxwatch = Conf.ReadInteger("watch", "maxentries", 0, true);
		if (!maxwatch)
			maxwatch = 32;
	}

	virtual int OnSetAway(User *user, const std::string &awaymsg)
	{
		std::string numeric;
		int inum;

		if (awaymsg.empty())
		{
			numeric = std::string(user->nick) + " " + user->ident + " " + user->dhost + " " + ConvToStr(ServerInstance->Time()) + " :is no longer away";
			inum = 599;
		}
		else
		{
			numeric = std::string(user->nick) + " " + user->ident + " " + user->dhost + " " + ConvToStr(ServerInstance->Time()) + " :" + awaymsg;
			inum = 598;
		}

		watchentries::iterator x = whos_watching_me->find(user->nick);
		if (x != whos_watching_me->end())
		{
			for (std::deque<User*>::iterator n = x->second.begin(); n != x->second.end(); n++)
			{
				if (!user->Visibility || user->Visibility->VisibleTo(user))
					(*n)->WriteNumeric(inum, numeric);
			}
		}

		return 0;
	}

	virtual void OnUserQuit(User* user, const std::string &reason, const std::string &oper_message)
	{
		watchentries::iterator x = whos_watching_me->find(user->nick);
		if (x != whos_watching_me->end())
		{
			for (std::deque<User*>::iterator n = x->second.begin(); n != x->second.end(); n++)
			{
				if (!user->Visibility || user->Visibility->VisibleTo(user))
					(*n)->WriteNumeric(601, "%s %s %s %s %lu :went offline", (*n)->nick ,user->nick, user->ident, user->dhost, ServerInstance->Time());

				watchlist* wl;
				if ((*n)->GetExt("watchlist", wl))
					/* We were on somebody's notify list, set ourselves offline */
					(*wl)[user->nick] = "";
			}
		}

		/* Now im quitting, if i have a notify list, im no longer watching anyone */
		watchlist* wl;
		if (user->GetExt("watchlist", wl))
		{
			/* Iterate every user on my watch list, and take me out of the whos_watching_me map for each one we're watching */
			for (watchlist::iterator i = wl->begin(); i != wl->end(); i++)
			{
				watchentries::iterator i2 = whos_watching_me->find(i->first);
				if (i2 != whos_watching_me->end())
				{
						/* People are watching this user, am i one of them? */
						std::deque<User*>::iterator n = std::find(i2->second.begin(), i2->second.end(), user);
						if (n != i2->second.end())
							/* I'm no longer watching you... */
							i2->second.erase(n);
	
						if (!i2->second.size())
							whos_watching_me->erase(user->nick);
				}
			}

			/* User's quitting, we're done with this. */
			delete wl;
			user->Shrink("watchlist");
		}
	}

	virtual void OnGarbageCollect()
	{
		watchentries* old_watch = whos_watching_me;
		whos_watching_me = new watchentries();

		for (watchentries::const_iterator n = old_watch->begin(); n != old_watch->end(); n++)
			whos_watching_me->insert(*n);

		delete old_watch;
	}

	virtual void OnCleanup(int target_type, void* item)
	{
		if (target_type == TYPE_USER)
		{
			watchlist* wl;
			User* user = (User*)item;

			if (user->GetExt("watchlist", wl))
			{
				user->Shrink("watchlist");
				delete wl;
			}
		}
	}

	virtual void OnPostConnect(User* user)
	{
		watchentries::iterator x = whos_watching_me->find(user->nick);
		if (x != whos_watching_me->end())
		{
			for (std::deque<User*>::iterator n = x->second.begin(); n != x->second.end(); n++)
			{
				if (!user->Visibility || user->Visibility->VisibleTo(user))
					(*n)->WriteNumeric(600, "%s %s %s %s %lu :arrived online", (*n)->nick, user->nick, user->ident, user->dhost, user->age);

				watchlist* wl;
				if ((*n)->GetExt("watchlist", wl))
					/* We were on somebody's notify list, set ourselves online */
					(*wl)[user->nick] = std::string(user->ident).append(" ").append(user->dhost).append(" ").append(ConvToStr(user->age));
			}
		}
	}

	virtual void OnUserPostNick(User* user, const std::string &oldnick)
	{
		watchentries::iterator new_offline = whos_watching_me->find(assign(oldnick));
		watchentries::iterator new_online = whos_watching_me->find(user->nick);

		if (new_offline != whos_watching_me->end())
		{
			for (std::deque<User*>::iterator n = new_offline->second.begin(); n != new_offline->second.end(); n++)
			{
				watchlist* wl;
				if ((*n)->GetExt("watchlist", wl))
				{
					if (!user->Visibility || user->Visibility->VisibleTo(user))
	 					(*n)->WriteNumeric(601, "%s %s %s %s %lu :went offline", (*n)->nick, oldnick.c_str(), user->ident, user->dhost, user->age);
					(*wl)[oldnick.c_str()] = "";
				}
			}
		}

		if (new_online != whos_watching_me->end())
		{
			for (std::deque<User*>::iterator n = new_online->second.begin(); n != new_online->second.end(); n++)
			{
				watchlist* wl;
				if ((*n)->GetExt("watchlist", wl))
				{
					(*wl)[user->nick] = std::string(user->ident).append(" ").append(user->dhost).append(" ").append(ConvToStr(user->age));
					if (!user->Visibility || user->Visibility->VisibleTo(user))
						(*n)->WriteNumeric(600, "%s %s %s :arrived online", (*n)->nick, user->nick, (*wl)[user->nick].c_str());
				}
			}
		}
	}	

	virtual void On005Numeric(std::string &output)
	{
		// we don't really have a limit...
		output = output + " WATCH=" + ConvToStr(maxwatch);
	}
	
	virtual ~Modulewatch()
	{
		delete whos_watching_me;
	}
	
	virtual Version GetVersion()
	{
		return Version(1, 2, 0, 0, VF_COMMON | VF_VENDOR, API_VERSION);
	}
};

MODULE_INIT(Modulewatch)

